/******************************************
 * @author kL <klk0@qq.com>
 * @date 2019/6/6
 * @doc 列表组件.
 * @end
 ******************************************/
const { ccclass, property, disallowMultiple, menu, executionOrder, requireComponent } = cc._decorator;

import ListItem from './ListItem';

enum TemplateType {
    NODE = 1,
    PREFAB = 2,
}

enum SlideType {
    NORMAL = 1,//普通
    ADHERING = 2,//粘附模式，将强制关闭滚动惯性
    PAGE = 3,//页面模式，将强制关闭滚动惯性
}

enum SelectedType {
    NONE = 0,
    SINGLE = 1,//单选
    MULT = 2,//多选
}

@ccclass
@disallowMultiple()
@menu('自定义组件/List')
@requireComponent(cc.ScrollView)
//脚本生命周期回调的执行优先级。小于 0 的脚本将优先执行，大于 0 的脚本将最后执行。该优先级只对 onLoad, onEnable, start, update 和 lateUpdate 有效，对 onDisable 和 onDestroy 无效。
@executionOrder(-5000)
export default class List extends cc.Component {
    //模板类型
    @property({ type: cc.Enum(TemplateType), tooltip: CC_DEV && '模板类型', })
    private templateType: TemplateType = TemplateType.NODE;
    //模板Item（Node）
    @property({
        type: cc.Node,
        tooltip: CC_DEV && '模板Item',
        visible() { return this.templateType == TemplateType.NODE; }
    })
    tmpNode: cc.Node = null;
    //模板Item（Prefab）
    @property({
        type: cc.Prefab,
        tooltip: CC_DEV && '模板Item',
        visible() { return this.templateType == TemplateType.PREFAB; }
    })
    tmpPrefab: cc.Prefab = null;
    // item脚本名字
    @property()
    itemComponentName: string = "ListItem";
    //滑动模式
    @property()
    private _slideMode: SlideType = SlideType.NORMAL;
    @property({
        type: cc.Enum(SlideType),
        tooltip: CC_DEV && '滑动模式'
    })
    set slideMode(val: SlideType) {
        this._slideMode = val;
    }
    get slideMode() {
        return this._slideMode;
    }
    //翻页作用距离
    @property({
        type: cc.Float,
        range: [0, 1, .1],
        tooltip: CC_DEV && '翻页作用距离',
        slide: true,
        visible() { return this._slideMode == SlideType.PAGE; }
    })
    public pageDistance: number = .3;
    //页面改变事件
    @property({
        type: cc.Component.EventHandler,
        tooltip: CC_DEV && '页面改变事件',
        visible() { return this._slideMode == SlideType.PAGE; }
    })
    private pageChangeEvent: cc.Component.EventHandler = new cc.Component.EventHandler();
    @property({
        tooltip: CC_DEV && '自适应尺寸（开启后每个item的尺寸固定后，会生成对应的坐标映射后续位置查找会直接查表而不是遍历计算。这个优化只针对纵向布局TOP_TO_BOTTOM这个分支，其他分支如果需要可以参考实现）',
    })
    adaptiveSize: boolean = false;
    //是否为虚拟列表（动态列表）
    @property()
    private _virtual: boolean = true;
    @property({
        type: cc.Boolean,
        tooltip: CC_DEV && '是否为虚拟列表（动态列表）'
    })
    set virtual(val: boolean) {
        if (val != null)
            this._virtual = val;
        if (!CC_DEV && this._numItems != 0) {
            this._onScrolling();
        }
    }
    get virtual() {
        return this._virtual;
    }
    //是否为循环列表
    @property({
        tooltip: CC_DEV && '是否为循环列表',
        visible() {
            let val: boolean = /*this.virtual &&*/ this.slideMode == SlideType.NORMAL;
            if (!val)
                this.cyclic = false;
            return val;
        }
    })
    public cyclic: boolean = false;
    //缺省居中
    @property({
        tooltip: CC_DEV && 'Item数量不足以填满Content时，是否居中显示Item（不支持Grid布局）',
        visible() { return this.virtual; }
    })
    public lackCenter: boolean = false;
    //缺省可滑动
    @property({
        tooltip: CC_DEV && 'Item数量不足以填满Content时，是否可滑动',
        visible() {
            let val: boolean = this.virtual && !this.lackCenter;
            if (!val)
                this.lackSlide = false;
            return val;
        }
    })
    public lackSlide: boolean = false;
    //刷新频率
    @property({ type: cc.Integer })
    private _updateRate: number = 0;
    @property({
        type: cc.Integer,
        range: [0, 6, 1],
        tooltip: CC_DEV && '刷新频率（值越大刷新频率越低、性能越高）',
        slide: true,
    })
    set updateRate(val: number) {
        if (val >= 0 && val <= 6) {
            this._updateRate = val;
        }
    }
    get updateRate() {
        return this._updateRate;
    }
    //分帧渲染（每帧渲染的Item数量（<=0时关闭分帧渲染））
    @property({
        type: cc.Integer,
        range: [0, 12, 1],
        tooltip: CC_DEV && '逐帧渲染时，每帧渲染的Item数量（<=0时关闭分帧渲染）',
        slide: true,
    })
    public frameByFrameRenderNum: number = 0;
    //渲染事件（渲染器）
    @property({
        type: cc.Component.EventHandler,
        tooltip: CC_DEV && '渲染事件（渲染器）',
    })
    private renderEvent: cc.Component.EventHandler = new cc.Component.EventHandler();
    //选择模式
    @property({
        type: cc.Enum(SelectedType),
        tooltip: CC_DEV && '选择模式'
    })
    public selectedMode: SelectedType = SelectedType.NONE;
    @property({
        tooltip: CC_DEV && '是否重复响应单选事件',
        visible() { return this.selectedMode == SelectedType.SINGLE; }
    })
    public repeatEventSingle: boolean = false;
    //触发选择事件
    @property({
        type: cc.Component.EventHandler,
        tooltip: CC_DEV && '触发选择事件',
        visible() { return this.selectedMode > SelectedType.NONE; }
    })
    private selectedEvent: cc.Component.EventHandler = new cc.Component.EventHandler();
    //当前选择id
    private _selectedId: number = -1;
    private _lastSelectedId: number;
    private multSelected: number[];
    set selectedId(val: number) {
        let item: any;
        switch (this.selectedMode) {
            case SelectedType.SINGLE: {
                if (!this.repeatEventSingle && val == this._selectedId)
                    return;
                item = this.getItemByListId(val);
                // if (!item && val >= 0)
                //     return;
                let listItem: ListItem;
                if (this._selectedId >= 0)
                    this._lastSelectedId = this._selectedId;
                else //如果＜0则取消选择，把_lastSelectedId也置空吧，如果以后有特殊需求再改吧。
                    this._lastSelectedId = null;
                this._selectedId = val;
                if (item) {
                    listItem = item.getComponent(this.itemComponentName);
                    listItem.selected = true;
                }
                if (this._lastSelectedId >= 0 && this._lastSelectedId != this._selectedId) {
                    let lastItem: any = this.getItemByListId(this._lastSelectedId);
                    if (lastItem) {
                        lastItem.getComponent(this.itemComponentName).selected = false;
                    }
                }
                if (this.selectedEvent) {
                    cc.Component.EventHandler.emitEvents([this.selectedEvent], item, val % this._actualNumItems, this._lastSelectedId == null ? null : (this._lastSelectedId % this._actualNumItems));
                }
                break;
            }
            case SelectedType.MULT: {
                item = this.getItemByListId(val);
                if (!item)
                    return;
                let listItem = item.getComponent(this.itemComponentName);
                if (this._selectedId >= 0)
                    this._lastSelectedId = this._selectedId;
                this._selectedId = val;
                let bool: boolean = !listItem.selected;
                listItem.selected = bool;
                let sub: number = this.multSelected.indexOf(val);
                if (bool && sub < 0) {
                    this.multSelected.push(val);
                } else if (!bool && sub >= 0) {
                    this.multSelected.splice(sub, 1);
                }
                if (this.selectedEvent) {
                    cc.Component.EventHandler.emitEvents([this.selectedEvent], item, val % this._actualNumItems, this._lastSelectedId == null ? null : (this._lastSelectedId % this._actualNumItems), bool);
                }
                break;
            }
        }
    }
    get selectedId() {
        return this._selectedId;
    }
    private _forceUpdate: boolean = false;
    private _align: number;
    private _horizontalDir: number;
    private _verticalDir: number;
    private _startAxis: number;
    private _alignCalcType: number;
    public content: cc.Node;
    private firstListId: number;
    public displayItemNum: number;
    private _updateDone: boolean = true;
    private _updateCounter: number;
    public _actualNumItems: number;
    private _cyclicNum: number;
    private _cyclicPos1: number;
    private _cyclicPos2: number;
    //列表数量
    @property({
        serializable: false
    })
    private _numItems: number = 0;
    set numItems(val: number) {
        if (!this.checkInited(false))
            return;
        if (val == null || val < 0) {
            cc.error('numItems set the wrong::', val);
            return;
        }
        this._actualNumItems = this._numItems = val;
        this._forceUpdate = true;
        this.resetAdaptiveSize();
        if (this._virtual) {
            this._resizeContent();
            if (this.cyclic) {
                this._numItems = this._cyclicNum * this._numItems;
            }
            this._onScrolling();
            if (!this.frameByFrameRenderNum && this.slideMode == SlideType.PAGE)
                this.curPageNum = this.nearestListId;
        } else {
            if (this.cyclic) {
                this._resizeContent();
                this._numItems = this._cyclicNum * this._numItems;
            }
            let layout: cc.Layout = this.content.getComponent(cc.Layout);
            if (layout) {
                layout.enabled = true;
            }
            this._delRedundantItem();

            this.firstListId = 0;
            if (this.frameByFrameRenderNum > 0) {
                //先渲染几个出来
                let len: number = this.frameByFrameRenderNum > this._numItems ? this._numItems : this.frameByFrameRenderNum;
                for (let n: number = 0; n < len; n++) {
                    this._createOrUpdateItem2(n);
                }
                if (this.frameByFrameRenderNum < this._numItems) {
                    this._updateCounter = this.frameByFrameRenderNum;
                    this._updateDone = false;
                }
            } else {
                for (let n: number = 0; n < this._numItems; n++) {
                    this._createOrUpdateItem2(n);
                }
                this.displayItemNum = this._numItems;
            }
        }
    }
    get numItems() {
        return this._actualNumItems;
    }

    private _inited: boolean = false;
    private _scrollView: cc.ScrollView;
    get scrollView() {
        return this._scrollView;
    }
    private _layout: cc.Layout;
    private _resizeMode: cc.Layout.ResizeMode;
    private _topGap: number;
    private _rightGap: number;
    private _bottomGap: number;
    private _leftGap: number;

    private _columnGap: number;
    private _lineGap: number;
    private _colLineNum: number;

    private _lastDisplayData: number[];
    public displayData: any[];
    private _pool: cc.NodePool;

    private _itemTmp: any;
    private _needUpdateWidget: boolean = false;
    private _itemSize: cc.Size;
    private _sizeType: boolean;

    public _customSize: any;
    public _customPos: number[] = [];

    private _fixDelta = 0;
    private frameCount: number;
    private _aniDelRuning: boolean = false;
    private _aniDelCB: Function;
    private _aniDelItem: any;
    private _aniDelBeforePos: cc.Vec2;
    private _aniDelBeforeScale: number;
    private viewTop: number;
    private viewRight: number;
    private viewBottom: number;
    private viewLeft: number;

    private _doneAfterUpdate: boolean = false;

    private elasticTop: number;
    private elasticRight: number;
    private elasticBottom: number;
    private elasticLeft: number;

    private scrollToListId: number;

    private adhering: boolean = false;

    private _adheringBarrier: boolean = false;
    private nearestListId: number;

    public curPageNum: number = 0;
    private _beganPos: number;
    private _scrollPos: number;
    private curScrollIsTouch: boolean;//当前滑动是否为手动

    private _scrollToListId: number;
    private _scrollToEndTime: number;
    private _scrollToSo: any;

    private _lack: boolean;
    private _allItemSize: number;
    private _allItemSizeNoEdge: number;
    private _cyclicAllItemSize: number;
    private _cycilcAllItemSizeNoEdge: number;

    private _scrollItem: any;//当前控制 ScrollView 滚动的 Item
    private _lastContentPosY: number = 0;

    //----------------------------------------------------------------------------

    onLoad() {
        this._init();
    }

    onDestroy() {
        if (cc.isValid(this._itemTmp))
            this._itemTmp.destroy();
        if (cc.isValid(this.tmpNode))
            this.tmpNode.destroy();
        this._pool && this._pool.clear();
    }

    private resetAdaptiveSize(): void {
        // item不定宽高模式下，如果新数据比老数据少时，同步到位置和尺寸数据。
        if (this.adaptiveSize) {
            if (this._numItems < this._customPos.length) {
                this._customPos.length = this._numItems;
            }
        }
        if (this._customSize) {
            const ids = Object.keys(this._customSize);
            const lastId = this._numItems - 1;
            for (let i = ids.length - 1; i >= 0; --i) {
                const id = ids[i];
                if (Number(id) > lastId) {
                    delete this._customSize[id];
                } else {
                    break;
                }
            }
        }
    }

    onEnable() {
        // if (!CC_EDITOR) 
        this._registerEvent();
        this._init();
        // 处理重新显示后，有可能上一次的动画移除还未播放完毕，导致动画卡住的问题
        if (this._aniDelRuning) {
            this._aniDelRuning = false;
            if (this._aniDelItem) {
                if (this._aniDelBeforePos) {
                    this._aniDelItem.position = this._aniDelBeforePos;
                    delete this._aniDelBeforePos;
                }
                if (this._aniDelBeforeScale) {
                    this._aniDelItem.scale = this._aniDelBeforeScale;
                    delete this._aniDelBeforeScale;
                }
                delete this._aniDelItem;
            }
            if (this._aniDelCB) {
                this._aniDelCB();
                delete this._aniDelCB;
            }
        }
    }

    onDisable() {
        // if (!CC_EDITOR) 
        this._unregisterEvent();
    }
    //注册事件
    _registerEvent() {
        this.node.on(cc.Node.EventType.TOUCH_START, this._onTouchStart, this, true);
        this.node.on('touch-up', this._onTouchUp, this);
        this.node.on(cc.Node.EventType.TOUCH_CANCEL, this._onTouchCancelled, this, true);
        this.node.on('scroll-began', this._onScrollBegan, this, true);
        this.node.on('scroll-ended', this._onScrollEnded, this, true);
        this.node.on('scrolling', this._onScrolling, this, true);
        this.node.on(cc.Node.EventType.SIZE_CHANGED, this._onSizeChanged, this);
    }
    //卸载事件
    _unregisterEvent() {
        this.node.off(cc.Node.EventType.TOUCH_START, this._onTouchStart, this, true);
        this.node.off('touch-up', this._onTouchUp, this);
        this.node.off(cc.Node.EventType.TOUCH_CANCEL, this._onTouchCancelled, this, true);
        this.node.off('scroll-began', this._onScrollBegan, this, true);
        this.node.off('scroll-ended', this._onScrollEnded, this, true);
        this.node.off('scrolling', this._onScrolling, this, true);
        this.node.off(cc.Node.EventType.SIZE_CHANGED, this._onSizeChanged, this);
    }
    //初始化各种..
    _init() {
        if (this._inited)
            return;

        this._scrollView = this.node.getComponent(cc.ScrollView);

        this.content = this._scrollView.content;
        if (!this.content) {
            cc.error(this.node.name + "'s cc.ScrollView unset content!");
            return;
        }

        this._layout = this.content.getComponent(cc.Layout);

        this._align = this._layout.type; //排列模式
        this._resizeMode = this._layout.resizeMode; //自适应模式
        this._startAxis = this._layout.startAxis;

        this._topGap = this._layout.paddingTop; //顶边距
        this._rightGap = this._layout.paddingRight; //右边距
        this._bottomGap = this._layout.paddingBottom; //底边距
        this._leftGap = this._layout.paddingLeft; //左边距

        this._columnGap = this._layout.spacingX; //列距
        this._lineGap = this._layout.spacingY; //行距

        this._colLineNum; //列数或行数（非GRID模式则=1，表示单列或单行）;

        this._verticalDir = this._layout.verticalDirection; //垂直排列子节点的方向
        this._horizontalDir = this._layout.horizontalDirection; //水平排列子节点的方向

        this.setTemplateItem(cc.instantiate(this.templateType == TemplateType.PREFAB ? this.tmpPrefab : this.tmpNode));

        // 特定的滑动模式处理
        if (this._slideMode == SlideType.ADHERING || this._slideMode == SlideType.PAGE) {
            this._scrollView.inertia = false;
            this._scrollView._onMouseWheel = function () {
                return;
            };
        }
        if (!this.virtual)         // lackCenter 仅支持 Virtual 模式
            this.lackCenter = false;

        this._lastDisplayData = []; //最后一次刷新的数据
        this.displayData = []; //当前数据
        this._pool = new cc.NodePool();    //这是个池子..
        this._forceUpdate = false;         //是否强制更新
        this._updateCounter = 0;           //当前分帧渲染帧数
        this._updateDone = true;           //分帧渲染是否完成

        this.curPageNum = 0;               //当前页数

        if (this.cyclic || 0) {
            this._scrollView._processAutoScrolling = this._processAutoScrolling.bind(this);
            this._scrollView._startBounceBackIfNeeded = function () {
                return false;
            };
            // t._scrollView._scrollChildren = function () {
            //     return false;
            // }
        }

        switch (this._align) {
            case cc.Layout.Type.HORIZONTAL: {
                switch (this._horizontalDir) {
                    case cc.Layout.HorizontalDirection.LEFT_TO_RIGHT:
                        this._alignCalcType = 1;
                        break;
                    case cc.Layout.HorizontalDirection.RIGHT_TO_LEFT:
                        this._alignCalcType = 2;
                        break;
                }
                break;
            }
            case cc.Layout.Type.VERTICAL: {
                switch (this._verticalDir) {
                    case cc.Layout.VerticalDirection.TOP_TO_BOTTOM:
                        this._alignCalcType = 3;
                        break;
                    case cc.Layout.VerticalDirection.BOTTOM_TO_TOP:
                        this._alignCalcType = 4;
                        break;
                }
                break;
            }
            case cc.Layout.Type.GRID: {
                switch (this._startAxis) {
                    case cc.Layout.AxisDirection.HORIZONTAL:
                        switch (this._verticalDir) {
                            case cc.Layout.VerticalDirection.TOP_TO_BOTTOM:
                                this._alignCalcType = 3;
                                break;
                            case cc.Layout.VerticalDirection.BOTTOM_TO_TOP:
                                this._alignCalcType = 4;
                                break;
                        }
                        break;
                    case cc.Layout.AxisDirection.VERTICAL:
                        switch (this._horizontalDir) {
                            case cc.Layout.HorizontalDirection.LEFT_TO_RIGHT:
                                this._alignCalcType = 1;
                                break;
                            case cc.Layout.HorizontalDirection.RIGHT_TO_LEFT:
                                this._alignCalcType = 2;
                                break;
                        }
                        break;
                }
                break;
            }
        }
        // 清空 content
        // t.content.children.forEach((child: cc.Node) => {
        //     child.removeFromParent();
        //     if (child != t.tmpNode && child.isValid)
        //         child.destroy();
        // });
        this.content.removeAllChildren();
        this._inited = true;
    }


    public reset() {
        if (!this._inited) {
            return;
        }
        this._selectedId = -1;
        this._lastSelectedId = undefined;
        this._lastDisplayData.length = 0; //最后一次刷新的数据
        this.displayData.length = 0; //当前数据
        this._pool.clear();    //这是个池子..
        this._customPos.length = 0;
        this._fixDelta = 0;
        this._forceUpdate = false;         //是否强制更新
        this._updateCounter = 0;           //当前分帧渲染帧数
        this._updateDone = true;           //分帧渲染是否完成
        this.curPageNum = 0;               //当前页数
        this.content.removeAllChildren();
        this._customSize = undefined;
    }

    /**
     * 为了实现循环列表，必须覆写cc.ScrollView的某些函数
     * @param {Number} dt
     */
    _processAutoScrolling(dt: number) {
        let brakingFactor: number = 1;
        this._scrollView['_autoScrollAccumulatedTime'] += dt * (1 / brakingFactor);

        let percentage: number = Math.min(1, this._scrollView['_autoScrollAccumulatedTime'] / this._scrollView['_autoScrollTotalTime']);
        if (this._scrollView['_autoScrollAttenuate']) {
            let time: number = percentage - 1;
            percentage = time * time * time * time * time + 1;
        }

        let newPosition: any = this._scrollView['_autoScrollStartPosition'].add(this._scrollView['_autoScrollTargetDelta'].mul(percentage));
        let EPSILON: number = this._scrollView['getScrollEndedEventTiming']();
        let reachedEnd: boolean = Math.abs(percentage - 1) <= EPSILON;

        let fireEvent: boolean = Math.abs(percentage - 1) <= this._scrollView['getScrollEndedEventTiming']();
        if (fireEvent && !this._scrollView['_isScrollEndedWithThresholdEventFired']) {
            this._scrollView['_dispatchEvent']('scroll-ended-with-threshold');
            this._scrollView['_isScrollEndedWithThresholdEventFired'] = true;
        }

        if (reachedEnd) {
            this._scrollView['_autoScrolling'] = false;
        }

        let deltaMove: any = newPosition.sub(this._scrollView.getContentPosition());
        this._scrollView['_moveContent'](this._scrollView['_clampDelta'](deltaMove), reachedEnd);
        this._scrollView['_dispatchEvent']('scrolling');

        // scollTo API controll move
        if (!this._scrollView['_autoScrolling']) {
            this._scrollView['_isBouncing'] = false;
            this._scrollView['_scrolling'] = false;
            this._scrollView['_dispatchEvent']('scroll-ended');
        }
    }
    //设置模板Item
    setTemplateItem(item: any) {
        if (!item)
            return;
        this._itemTmp = item;

        if (this._resizeMode == cc.Layout.ResizeMode.CHILDREN)
            this._itemSize = this._layout.cellSize;
        else
            this._itemSize = cc.size(item.width, item.height);

        //获取ListItem，如果没有就取消选择模式
        let com = item.getComponent(this.itemComponentName);
        let remove = false;
        if (!com)
            remove = true;
        // if (com) {
        //     if (!com._btnCom && !item.getComponent(cc.Button)) {
        //         remove = true;
        //     }
        // }
        if (remove) {
            this.selectedMode = SelectedType.NONE;
        }
        com = item.getComponent(cc.Widget);
        if (com && com.enabled) {
            this._needUpdateWidget = true;
        }
        if (this.selectedMode == SelectedType.MULT)
            this.multSelected = [];

        switch (this._align) {
            case cc.Layout.Type.HORIZONTAL:
                this._colLineNum = 1;
                this._sizeType = false;
                break;
            case cc.Layout.Type.VERTICAL:
                this._colLineNum = 1;
                this._sizeType = true;
                break;
            case cc.Layout.Type.GRID:
                switch (this._startAxis) {
                    case cc.Layout.AxisDirection.HORIZONTAL:
                        //计算列数
                        let trimW: number = this.content.width - this._leftGap - this._rightGap;
                        this._colLineNum = Math.floor((trimW + this._columnGap) / (this._itemSize.width + this._columnGap));
                        this._sizeType = true;
                        break;
                    case cc.Layout.AxisDirection.VERTICAL:
                        //计算行数
                        let trimH: number = this.content.height - this._topGap - this._bottomGap;
                        this._colLineNum = Math.floor((trimH + this._lineGap) / (this._itemSize.height + this._lineGap));
                        this._sizeType = false;
                        break;
                }
                break;
        }
    }
    /**
     * 检查是否初始化
     * @param {Boolean} printLog 是否打印错误信息
     * @returns
     */
    checkInited(printLog: boolean = true) {
        if (!this._inited) {
            if (printLog)
                cc.error('List initialization not completed!');
            return false;
        }
        return true;
    }
    //禁用 Layout 组件，自行计算 Content Size
    _resizeContent() {
        let result: number;
        switch (this._align) {
            case cc.Layout.Type.HORIZONTAL: {
                if (this._customSize) {
                    let fixed: any = this._getFixedSize(null);
                    result = this._leftGap + fixed.val + (this._itemSize.width * (this._numItems - fixed.count)) + (this._columnGap * (this._numItems - 1)) + this._rightGap;
                } else {
                    result = this._leftGap + (this._itemSize.width * this._numItems) + (this._columnGap * (this._numItems - 1)) + this._rightGap;
                }
                break;
            }
            case cc.Layout.Type.VERTICAL: {
                if (this._customSize) {
                    let fixed: any = this._getFixedSize(null);
                    const fixedDelta = this.adaptiveSize ? this._fixDelta : 0;
                    result = this._topGap + fixed.val + (this._itemSize.height * (this._numItems - fixed.count)) + (this._lineGap * (this._numItems - 1)) + this._bottomGap + fixedDelta;
                } else {
                    result = this._topGap + (this._itemSize.height * this._numItems) + (this._lineGap * (this._numItems - 1)) + this._bottomGap;
                }
                break;
            }
            case cc.Layout.Type.GRID: {
                //网格模式不支持居中
                if (this.lackCenter)
                    this.lackCenter = false;
                switch (this._startAxis) {
                    case cc.Layout.AxisDirection.HORIZONTAL:
                        let lineNum: number = Math.ceil(this._numItems / this._colLineNum);
                        result = this._topGap + (this._itemSize.height * lineNum) + (this._lineGap * (lineNum - 1)) + this._bottomGap;
                        break;
                    case cc.Layout.AxisDirection.VERTICAL:
                        let colNum: number = Math.ceil(this._numItems / this._colLineNum);
                        result = this._leftGap + (this._itemSize.width * colNum) + (this._columnGap * (colNum - 1)) + this._rightGap;
                        break;
                }
                break;
            }
        }

        let layout: cc.Layout = this.content.getComponent(cc.Layout);
        if (layout)
            layout.enabled = false;

        this._allItemSize = result;
        this._allItemSizeNoEdge = this._allItemSize - (this._sizeType ? (this._topGap + this._bottomGap) : (this._leftGap + this._rightGap));

        if (this.cyclic) {
            let totalSize: number = (this._sizeType ? this.node.height : this.node.width);

            this._cyclicPos1 = 0;
            totalSize -= this._cyclicPos1;
            this._cyclicNum = Math.ceil(totalSize / this._allItemSizeNoEdge) + 1;
            let spacing: number = this._sizeType ? this._lineGap : this._columnGap;
            this._cyclicPos2 = this._cyclicPos1 + this._allItemSizeNoEdge + spacing;
            this._cyclicAllItemSize = this._allItemSize + (this._allItemSizeNoEdge * (this._cyclicNum - 1)) + (spacing * (this._cyclicNum - 1));
            this._cycilcAllItemSizeNoEdge = this._allItemSizeNoEdge * this._cyclicNum;
            this._cycilcAllItemSizeNoEdge += spacing * (this._cyclicNum - 1);
            // cc.log('_cyclicNum ->', t._cyclicNum, t._allItemSizeNoEdge, t._allItemSize, t._cyclicPos1, t._cyclicPos2);
        }

        this._lack = !this.cyclic && this._allItemSize < (this._sizeType ? this.node.height : this.node.width);
        let slideOffset: number = ((!this._lack || !this.lackCenter) && this.lackSlide) ? 0 : .1;

        let targetWH: number = this._lack ? ((this._sizeType ? this.node.height : this.node.width) - slideOffset) : (this.cyclic ? this._cyclicAllItemSize : this._allItemSize);
        if (targetWH < 0)
            targetWH = 0;

        if (this._sizeType) {
            this.content.height = targetWH;
        } else {
            this.content.width = targetWH;
        }

        cc.log('_resizeContent()  numItems =', this._numItems, '，content =', this.content, ',cH = ', this.content.height);
    }

    //滚动进行时...
    _onScrolling(ev: cc.Event = null) {
        let directionType: 'up' | 'down';
        if (this.adaptiveSize) {
            const direction = this.content.y - this._lastContentPosY;
            if (direction > 0) {
                directionType = 'up';
            } else if (direction < 0) {
                directionType = 'down';
            } else {
                directionType = 'up';
            }
            this._lastContentPosY = this.content.y;
        }
        if (this.frameCount == null)
            this.frameCount = this._updateRate;
        if (!this._forceUpdate && (ev && ev.type != 'scroll-ended') && this.frameCount > 0) {
            this.frameCount--;
            return;
        } else
            this.frameCount = this._updateRate;

        if (this._aniDelRuning)
            return;

        //循环列表处理
        if (this.cyclic) {
            let scrollPos: any = this.content.getPosition();
            scrollPos = this._sizeType ? scrollPos.y : scrollPos.x;

            let addVal = this._allItemSizeNoEdge + (this._sizeType ? this._lineGap : this._columnGap);
            let add: any = this._sizeType ? cc.v2(0, addVal) : cc.v2(addVal, 0);

            switch (this._alignCalcType) {
                case 1://单行HORIZONTAL（LEFT_TO_RIGHT）、网格VERTICAL（LEFT_TO_RIGHT）
                    if (scrollPos > -this._cyclicPos1) {
                        this.content.x = -this._cyclicPos2;
                        if (this._scrollView.isAutoScrolling()) {
                            this._scrollView['_autoScrollStartPosition'] = this._scrollView['_autoScrollStartPosition'].sub(add);
                        }
                        // if (this._beganPos) {
                        //     this._beganPos += add;
                        // }
                    } else if (scrollPos < -this._cyclicPos2) {
                        this.content.x = -this._cyclicPos1;
                        if (this._scrollView.isAutoScrolling()) {
                            this._scrollView['_autoScrollStartPosition'] = this._scrollView['_autoScrollStartPosition'].add(add);
                        }
                        // if (this._beganPos) {
                        //     this._beganPos -= add;
                        // }
                    }
                    break;
                case 2://单行HORIZONTAL（RIGHT_TO_LEFT）、网格VERTICAL（RIGHT_TO_LEFT）
                    if (scrollPos < this._cyclicPos1) {
                        this.content.x = this._cyclicPos2;
                        if (this._scrollView.isAutoScrolling()) {
                            this._scrollView['_autoScrollStartPosition'] = this._scrollView['_autoScrollStartPosition'].add(add);
                        }
                    } else if (scrollPos > this._cyclicPos2) {
                        this.content.x = this._cyclicPos1;
                        if (this._scrollView.isAutoScrolling()) {
                            this._scrollView['_autoScrollStartPosition'] = this._scrollView['_autoScrollStartPosition'].sub(add);
                        }
                    }
                    break;
                case 3://单列VERTICAL（TOP_TO_BOTTOM）、网格HORIZONTAL（TOP_TO_BOTTOM）
                    if (scrollPos < this._cyclicPos1) {
                        this.content.y = this._cyclicPos2;
                        if (this._scrollView.isAutoScrolling()) {
                            this._scrollView['_autoScrollStartPosition'] = this._scrollView['_autoScrollStartPosition'].add(add);
                        }
                    } else if (scrollPos > this._cyclicPos2) {
                        this.content.y = this._cyclicPos1;
                        if (this._scrollView.isAutoScrolling()) {
                            this._scrollView['_autoScrollStartPosition'] = this._scrollView['_autoScrollStartPosition'].sub(add);
                        }
                    }
                    break;
                case 4://单列VERTICAL（BOTTOM_TO_TOP）、网格HORIZONTAL（BOTTOM_TO_TOP）
                    if (scrollPos > -this._cyclicPos1) {
                        this.content.y = -this._cyclicPos2;
                        if (this._scrollView.isAutoScrolling()) {
                            this._scrollView['_autoScrollStartPosition'] = this._scrollView['_autoScrollStartPosition'].sub(add);
                        }
                    } else if (scrollPos < -this._cyclicPos2) {
                        this.content.y = -this._cyclicPos1;
                        if (this._scrollView.isAutoScrolling()) {
                            this._scrollView['_autoScrollStartPosition'] = this._scrollView['_autoScrollStartPosition'].add(add);
                        }
                    }
                    break;
            }
        }

        this._calcViewPos();

        let vTop: number, vRight: number, vBottom: number, vLeft: number;
        if (this._sizeType) {
            vTop = this.viewTop;
            vBottom = this.viewBottom;
        } else {
            vRight = this.viewRight;
            vLeft = this.viewLeft;
        }

        if (this._virtual) {
            this.displayData = [];
            let itemPos: any;
            let curId: number = 0;
            let endId: number = this._numItems - 1;
            if (this._customSize) {
                let breakFor: boolean = false;
                //如果该item的位置在可视区域内，就推入displayData
                if (!this.adaptiveSize) {
                    for (; curId <= endId && !breakFor; curId++) {
                        itemPos = this._calcItemPos(curId);
                        switch (this._align) {
                            case cc.Layout.Type.HORIZONTAL:
                                if (itemPos.right >= vLeft && itemPos.left <= vRight) {
                                    this.displayData.push(itemPos);
                                } else if (curId != 0 && this.displayData.length > 0) {
                                    breakFor = true;
                                }
                                break;
                            case cc.Layout.Type.VERTICAL:
                                if (itemPos.bottom <= vTop && itemPos.top >= vBottom) {
                                    this.displayData.push(itemPos);
                                } else if (curId != 0 && this.displayData.length > 0) {
                                    breakFor = true;
                                }
                                break;
                            case cc.Layout.Type.GRID:
                                switch (this._startAxis) {
                                    case cc.Layout.AxisDirection.HORIZONTAL:
                                        if (itemPos.bottom <= vTop && itemPos.top >= vBottom) {
                                            this.displayData.push(itemPos);
                                        } else if (curId != 0 && this.displayData.length > 0) {
                                            breakFor = true;
                                        }
                                        break;
                                    case cc.Layout.AxisDirection.VERTICAL:
                                        if (itemPos.right >= vLeft && itemPos.left <= vRight) {
                                            this.displayData.push(itemPos);
                                        } else if (curId != 0 && this.displayData.length > 0) {
                                            breakFor = true;
                                        }
                                        break;
                                }
                                break;
                        }
                    }
                } else {
                    if (directionType === 'up') {
                        curId = this._lastDisplayData[0] ? this._lastDisplayData[0] : curId;
                        for (; curId <= endId && !breakFor; curId++) {
                            itemPos = this._calcItemPos(curId);
                            switch (this._align) {
                                case cc.Layout.Type.VERTICAL:
                                    if (itemPos.bottom <= vTop && itemPos.top >= vBottom) {
                                        this.displayData.push(itemPos);
                                    } else if (curId != 0 && this.displayData.length > 0) {
                                        breakFor = true;
                                    }
                                    break;
                            }
                        }
                    } else {
                        curId = this._lastDisplayData[this._lastDisplayData.length - 1] ? this._lastDisplayData[this._lastDisplayData.length - 1] : this._numItems - 1;
                        endId = 0;
                        for (; curId >= endId && !breakFor; curId--) {
                            itemPos = this._calcItemPos(curId);
                            switch (this._align) {
                                case cc.Layout.Type.VERTICAL:
                                    if (itemPos.bottom <= vTop && itemPos.top >= vBottom) {
                                        this.displayData.push(itemPos);
                                    } else if (curId != 0 && this.displayData.length > 0) {
                                        breakFor = true;
                                    }
                                    break;
                            }
                        }
                        this.displayData.sort((a, b) => { return a.id - b.id; });
                    }
                }

            } else {
                let ww: number = this._itemSize.width + this._columnGap;
                let hh: number = this._itemSize.height + this._lineGap;
                switch (this._alignCalcType) {
                    case 1://单行HORIZONTAL（LEFT_TO_RIGHT）、网格VERTICAL（LEFT_TO_RIGHT）
                        curId = (vLeft - this._leftGap) / ww;
                        endId = (vRight - this._leftGap) / ww;
                        break;
                    case 2://单行HORIZONTAL（RIGHT_TO_LEFT）、网格VERTICAL（RIGHT_TO_LEFT）
                        curId = (-vRight - this._rightGap) / ww;
                        endId = (-vLeft - this._rightGap) / ww;
                        break;
                    case 3://单列VERTICAL（TOP_TO_BOTTOM）、网格HORIZONTAL（TOP_TO_BOTTOM）
                        curId = (-vTop - this._topGap) / hh;
                        endId = (-vBottom - this._topGap) / hh;
                        break;
                    case 4://单列VERTICAL（BOTTOM_TO_TOP）、网格HORIZONTAL（BOTTOM_TO_TOP）
                        curId = (vBottom - this._bottomGap) / hh;
                        endId = (vTop - this._bottomGap) / hh;
                        break;
                }
                curId = Math.floor(curId) * this._colLineNum;
                endId = Math.ceil(endId) * this._colLineNum;
                endId--;
                if (curId < 0)
                    curId = 0;
                if (endId >= this._numItems)
                    endId = this._numItems - 1;
                for (; curId <= endId; curId++) {
                    this.displayData.push(this._calcItemPos(curId));
                }
            }
            this._delRedundantItem();
            if (this.displayData.length <= 0 || !this._numItems) { //if none, delete all.
                this._lastDisplayData = [];
                return;
            }
            this.firstListId = this.displayData[0].id;
            this.displayItemNum = this.displayData.length;

            let len: number = this._lastDisplayData.length;

            let haveDataChange: boolean = this.displayItemNum != len;
            if (haveDataChange) {
                // 如果是逐帧渲染，需要排序
                if (this.frameByFrameRenderNum > 0) {
                    this._lastDisplayData.sort((a, b) => { return a - b; });
                }
                // 因List的显示数据是有序的，所以只需要判断数组长度是否相等，以及头、尾两个元素是否相等即可。
                haveDataChange = this.firstListId != this._lastDisplayData[0] || this.displayData[this.displayItemNum - 1].id != this._lastDisplayData[len - 1];
            }

            if (this._forceUpdate || haveDataChange) {    //如果是强制更新
                if (this.frameByFrameRenderNum > 0) {
                    // if (this._updateDone) {
                    // this._lastDisplayData = [];
                    //逐帧渲染
                    if (this._numItems > 0) {
                        if (!this._updateDone) {
                            this._doneAfterUpdate = true;
                        } else {
                            this._updateCounter = 0;
                        }
                        this._updateDone = false;
                    } else {
                        this._updateCounter = 0;
                        this._updateDone = true;
                    }
                    // }
                } else {
                    //直接渲染
                    this._lastDisplayData = [];
                    // cc.log('List Display Data II::', this.displayData);
                    for (let c = 0; c < this.displayItemNum; c++) {
                        this._createOrUpdateItem(this.displayData[c]);
                    }
                    this._forceUpdate = false;
                }
            }
            this._calcNearestItem();
        }
    }
    //计算可视范围
    _calcViewPos() {
        let scrollPos: any = this.content.getPosition();
        switch (this._alignCalcType) {
            case 1://单行HORIZONTAL（LEFT_TO_RIGHT）、网格VERTICAL（LEFT_TO_RIGHT）
                this.elasticLeft = scrollPos.x > 0 ? scrollPos.x : 0;
                this.viewLeft = (scrollPos.x < 0 ? -scrollPos.x : 0) - this.elasticLeft;
                this.viewRight = this.viewLeft + this.node.width;
                this.elasticRight = this.viewRight > this.content.width ? Math.abs(this.viewRight - this.content.width) : 0;
                this.viewRight += this.elasticRight;
                // cc.log(this.elasticLeft, this.elasticRight, this.viewLeft, this.viewRight);
                break;
            case 2://单行HORIZONTAL（RIGHT_TO_LEFT）、网格VERTICAL（RIGHT_TO_LEFT）
                this.elasticRight = scrollPos.x < 0 ? -scrollPos.x : 0;
                this.viewRight = (scrollPos.x > 0 ? -scrollPos.x : 0) + this.elasticRight;
                this.viewLeft = this.viewRight - this.node.width;
                this.elasticLeft = this.viewLeft < -this.content.width ? Math.abs(this.viewLeft + this.content.width) : 0;
                this.viewLeft -= this.elasticLeft;
                // cc.log(this.elasticLeft, this.elasticRight, this.viewLeft, this.viewRight);
                break;
            case 3://单列VERTICAL（TOP_TO_BOTTOM）、网格HORIZONTAL（TOP_TO_BOTTOM）
                this.elasticTop = scrollPos.y < 0 ? Math.abs(scrollPos.y) : 0;
                this.viewTop = (scrollPos.y > 0 ? -scrollPos.y : 0) + this.elasticTop;
                this.viewBottom = this.viewTop - this.node.height;
                this.elasticBottom = this.viewBottom < -this.content.height ? Math.abs(this.viewBottom + this.content.height) : 0;
                this.viewBottom += this.elasticBottom;
                // cc.log(this.elasticTop, this.elasticBottom, this.viewTop, this.viewBottom);
                break;
            case 4://单列VERTICAL（BOTTOM_TO_TOP）、网格HORIZONTAL（BOTTOM_TO_TOP）
                this.elasticBottom = scrollPos.y > 0 ? Math.abs(scrollPos.y) : 0;
                this.viewBottom = (scrollPos.y < 0 ? -scrollPos.y : 0) - this.elasticBottom;
                this.viewTop = this.viewBottom + this.node.height;
                this.elasticTop = this.viewTop > this.content.height ? Math.abs(this.viewTop - this.content.height) : 0;
                this.viewTop -= this.elasticTop;
                // cc.log(this.elasticTop, this.elasticBottom, this.viewTop, this.viewBottom);
                break;
        }
    }
    //计算位置 根据id
    _calcItemPos(id: number) {
        let width: number, height: number, top: number, bottom: number, left: number, right: number, itemX: number, itemY: number;
        switch (this._align) {
            case cc.Layout.Type.HORIZONTAL:
                switch (this._horizontalDir) {
                    case cc.Layout.HorizontalDirection.LEFT_TO_RIGHT: {
                        if (this._customSize) {
                            let fixed: any = this._getFixedSize(id);
                            left = this._leftGap + ((this._itemSize.width + this._columnGap) * (id - fixed.count)) + (fixed.val + (this._columnGap * fixed.count));
                            let cs: number = this._customSize[id];
                            width = (cs > 0 ? cs : this._itemSize.width);
                        } else {
                            left = this._leftGap + ((this._itemSize.width + this._columnGap) * id);
                            width = this._itemSize.width;
                        }
                        if (this.lackCenter) {
                            left -= this._leftGap;
                            let offset: number = (this.content.width / 2) - (this._allItemSizeNoEdge / 2);
                            left += offset;
                        }
                        right = left + width;
                        return {
                            id: id,
                            left: left,
                            right: right,
                            x: left + (this._itemTmp.anchorX * width),
                            y: this._itemTmp.y,
                        };
                    }
                    case cc.Layout.HorizontalDirection.RIGHT_TO_LEFT: {
                        if (this._customSize) {
                            let fixed: any = this._getFixedSize(id);
                            right = -this._rightGap - ((this._itemSize.width + this._columnGap) * (id - fixed.count)) - (fixed.val + (this._columnGap * fixed.count));
                            let cs: number = this._customSize[id];
                            width = (cs > 0 ? cs : this._itemSize.width);
                        } else {
                            right = -this._rightGap - ((this._itemSize.width + this._columnGap) * id);
                            width = this._itemSize.width;
                        }
                        if (this.lackCenter) {
                            right += this._rightGap;
                            let offset: number = (this.content.width / 2) - (this._allItemSizeNoEdge / 2);
                            right -= offset;
                        }
                        left = right - width;
                        return {
                            id: id,
                            right: right,
                            left: left,
                            x: left + (this._itemTmp.anchorX * width),
                            y: this._itemTmp.y,
                        };
                    }
                }
                break;
            case cc.Layout.Type.VERTICAL: {
                switch (this._verticalDir) {
                    case cc.Layout.VerticalDirection.TOP_TO_BOTTOM: {
                        if (this._customSize) {
                            let cs: number = this._customSize[id];
                            height = (cs > 0 ? cs : this._itemSize.height);
                            if (this.adaptiveSize) {
                                if (this._customPos[id] != null) {
                                    top = this._customPos[id];
                                    // cc.warn('1');
                                } else if (this._customPos[id - 1] != null) {
                                    const preItemHeight = this._customSize[id - 1] ? this._customSize[id - 1] : this._itemSize.height;
                                    top = this._customPos[id - 1] - preItemHeight - this._lineGap;
                                    this._customPos[id] = top;
                                    // cc.warn('2');
                                } else {
                                    let fixed: any = this._getFixedSize(id);
                                    top = -this._topGap - ((this._itemSize.height + this._lineGap) * (id - fixed.count)) - (fixed.val + (this._lineGap * fixed.count));
                                    this._customPos[id] = top;
                                    // cc.warn('3');
                                }
                            } else {
                                let fixed: any = this._getFixedSize(id);
                                top = -this._topGap - ((this._itemSize.height + this._lineGap) * (id - fixed.count)) - (fixed.val + (this._lineGap * fixed.count));
                            }
                        } else {
                            top = -this._topGap - ((this._itemSize.height + this._lineGap) * id);
                            height = this._itemSize.height;
                        }
                        if (this.lackCenter) {
                            top += this._topGap;
                            let offset: number = (this.content.height / 2) - (this._allItemSizeNoEdge / 2);
                            top -= offset;
                        }
                        bottom = top - height;
                        return {
                            id: id,
                            top: top,
                            bottom: bottom,
                            x: this._itemTmp.x,
                            y: bottom + (this._itemTmp.anchorY * height),
                        };
                    }
                    case cc.Layout.VerticalDirection.BOTTOM_TO_TOP: {
                        if (this._customSize) {
                            let fixed: any = this._getFixedSize(id);
                            bottom = this._bottomGap + ((this._itemSize.height + this._lineGap) * (id - fixed.count)) + (fixed.val + (this._lineGap * fixed.count));
                            let cs: number = this._customSize[id];
                            height = (cs > 0 ? cs : this._itemSize.height);
                        } else {
                            bottom = this._bottomGap + ((this._itemSize.height + this._lineGap) * id);
                            height = this._itemSize.height;
                        }
                        if (this.lackCenter) {
                            bottom -= this._bottomGap;
                            let offset: number = (this.content.height / 2) - (this._allItemSizeNoEdge / 2);
                            bottom += offset;
                        }
                        top = bottom + height;
                        return {
                            id: id,
                            top: top,
                            bottom: bottom,
                            x: this._itemTmp.x,
                            y: bottom + (this._itemTmp.anchorY * height),
                        };
                        break;
                    }
                }
            }
            case cc.Layout.Type.GRID: {
                let colLine: number = Math.floor(id / this._colLineNum);
                switch (this._startAxis) {
                    case cc.Layout.AxisDirection.HORIZONTAL: {
                        switch (this._verticalDir) {
                            case cc.Layout.VerticalDirection.TOP_TO_BOTTOM: {
                                top = -this._topGap - ((this._itemSize.height + this._lineGap) * colLine);
                                bottom = top - this._itemSize.height;
                                itemY = bottom + (this._itemTmp.anchorY * this._itemSize.height);
                                break;
                            }
                            case cc.Layout.VerticalDirection.BOTTOM_TO_TOP: {
                                bottom = this._bottomGap + ((this._itemSize.height + this._lineGap) * colLine);
                                top = bottom + this._itemSize.height;
                                itemY = bottom + (this._itemTmp.anchorY * this._itemSize.height);
                                break;
                            }
                        }
                        itemX = this._leftGap + ((id % this._colLineNum) * (this._itemSize.width + this._columnGap));
                        switch (this._horizontalDir) {
                            case cc.Layout.HorizontalDirection.LEFT_TO_RIGHT: {
                                itemX += (this._itemTmp.anchorX * this._itemSize.width);
                                itemX -= (this.content.anchorX * this.content.width);
                                break;
                            }
                            case cc.Layout.HorizontalDirection.RIGHT_TO_LEFT: {
                                itemX += ((1 - this._itemTmp.anchorX) * this._itemSize.width);
                                itemX -= ((1 - this.content.anchorX) * this.content.width);
                                itemX *= -1;
                                break;
                            }
                        }
                        return {
                            id: id,
                            top: top,
                            bottom: bottom,
                            x: itemX,
                            y: itemY,
                        };
                    }
                    case cc.Layout.AxisDirection.VERTICAL: {
                        switch (this._horizontalDir) {
                            case cc.Layout.HorizontalDirection.LEFT_TO_RIGHT: {
                                left = this._leftGap + ((this._itemSize.width + this._columnGap) * colLine);
                                right = left + this._itemSize.width;
                                itemX = left + (this._itemTmp.anchorX * this._itemSize.width);
                                itemX -= (this.content.anchorX * this.content.width);
                                break;
                            }
                            case cc.Layout.HorizontalDirection.RIGHT_TO_LEFT: {
                                right = -this._rightGap - ((this._itemSize.width + this._columnGap) * colLine);
                                left = right - this._itemSize.width;
                                itemX = left + (this._itemTmp.anchorX * this._itemSize.width);
                                itemX += ((1 - this.content.anchorX) * this.content.width);
                                break;
                            }
                        }
                        itemY = -this._topGap - ((id % this._colLineNum) * (this._itemSize.height + this._lineGap));
                        switch (this._verticalDir) {
                            case cc.Layout.VerticalDirection.TOP_TO_BOTTOM: {
                                itemY -= ((1 - this._itemTmp.anchorY) * this._itemSize.height);
                                itemY += ((1 - this.content.anchorY) * this.content.height);
                                break;
                            }
                            case cc.Layout.VerticalDirection.BOTTOM_TO_TOP: {
                                itemY -= ((this._itemTmp.anchorY) * this._itemSize.height);
                                itemY += (this.content.anchorY * this.content.height);
                                itemY *= -1;
                                break;
                            }
                        }
                        return {
                            id: id,
                            left: left,
                            right: right,
                            x: itemX,
                            y: itemY,
                        };
                    }
                }
                break;
            }
        }
    }
    //计算已存在的Item的位置
    _calcExistItemPos(id: number) {
        let item: any = this.getItemByListId(id);
        if (!item)
            return null;
        let data: any = {
            id: id,
            x: item.x,
            y: item.y,
        };
        if (this._sizeType) {
            data.top = item.y + (item.height * (1 - item.anchorY));
            data.bottom = item.y - (item.height * item.anchorY);
        } else {
            data.left = item.x - (item.width * item.anchorX);
            data.right = item.x + (item.width * (1 - item.anchorX));
        }
        return data;
    }
    //获取Item位置
    getItemPos(id: number) {
        if (this._virtual)
            return this._calcItemPos(id);
        else {
            if (this.frameByFrameRenderNum)
                return this._calcItemPos(id);
            else
                return this._calcExistItemPos(id);
        }
    }
    //获取固定尺寸
    _getFixedSize(listId: number) {
        if (!this._customSize)
            return null;
        if (listId == null)
            listId = this._numItems;
        let fixed: number = 0;
        let count: number = 0;
        if (!this.adaptiveSize) {
            for (let id in this._customSize) {
                if (parseInt(id) < listId) {
                    fixed += this._customSize[id];
                    count++;
                }
            }
        }
        return {
            val: fixed,
            count: count,
        };
    }
    //滚动结束时..
    _onScrollBegan() {
        this._beganPos = this._sizeType ? this.viewTop : this.viewLeft;
    }
    //滚动结束时..
    _onScrollEnded() {
        this.curScrollIsTouch = false;
        if (this.scrollToListId != null) {
            let item: any = this.getItemByListId(this.scrollToListId);
            this.scrollToListId = null;
            if (item) {
                cc.tween(item)
                    .to(.1, { scale: 1.06 })
                    .to(.1, { scale: 1 })
                    .start();
            }
        }
        this._onScrolling();

        if (this._slideMode == SlideType.ADHERING &&
            !this.adhering
        ) {
            //cc.log(t.adhering, t._scrollView.isAutoScrolling(), t._scrollView.isScrolling());
            this.adhere();
        } else if (this._slideMode == SlideType.PAGE) {
            if (this._beganPos != null && this.curScrollIsTouch) {
                this._pageAdhere();
            } else {
                this.adhere();
            }
        } else {
            this._calcNearestItem();
        }
        this.onEndEvent();
    }

    //列表滑动到底的事件
    private _endEvent: Function;
    public setEndEvent(event: Function) {
        this._endEvent = event;
    }
    private onEndEvent() {
        // this._onScrolling();
        if (this.nearestListId == this._numItems - 1) {
            // cc.log("到底了")
            if (this._endEvent) {
                this._endEvent();
            }
        }
    }

    public isBottom(): boolean {
        const temp = this.content.height - this.node.height;
        if (temp <= 0) {
            return true;
        } else {
            return Math.abs(this.content.getPosition().y - temp) <= 20;
        }
    }

    // 触摸时
    _onTouchStart(ev, captureListeners) {
        if (this._scrollView['hasNestedViewGroup'](ev, captureListeners))
            return;
        this.curScrollIsTouch = true;
        let isMe = ev.eventPhase === cc.Event.AT_TARGET && ev.target === this.node;
        if (!isMe) {
            let itemNode: any = ev.target;
            while (itemNode._listId == null && itemNode.parent)
                itemNode = itemNode.parent;
            this._scrollItem = itemNode._listId != null ? itemNode : ev.target;
        }
    }
    //触摸抬起时..
    _onTouchUp() {
        this._scrollPos = null;
        if (this._slideMode == SlideType.ADHERING) {
            if (this.adhering)
                this._adheringBarrier = true;
            this.adhere();
        } else if (this._slideMode == SlideType.PAGE) {
            if (this._beganPos != null) {
                this._pageAdhere();
            } else {
                this.adhere();
            }
        }
        this._scrollItem = null;
    }

    _onTouchCancelled(ev, captureListeners) {
        if (this._scrollView['hasNestedViewGroup'](ev, captureListeners) || ev.simulate)
            return;

        this._scrollPos = null;
        if (this._slideMode == SlideType.ADHERING) {
            if (this.adhering)
                this._adheringBarrier = true;
            this.adhere();
        } else if (this._slideMode == SlideType.PAGE) {
            if (this._beganPos != null) {
                this._pageAdhere();
            } else {
                this.adhere();
            }
        }
        this._scrollItem = null;
    }
    //当尺寸改变
    _onSizeChanged() {
        if (this.checkInited(false))
            this._onScrolling();
    }
    //当Item自适应
    _onItemAdaptive(item: cc.Node): void {
        // if (this.checkInited(false)) {
        let defaultHeight = 0;
        let defaultWidth = 0;
        if (this._sizeType) {
            defaultHeight = this._customSize && this._customSize[item._listId] ? this._customSize[item._listId] : this._itemSize.height;
        } else {
            defaultWidth = this._customSize && this._customSize[item._listId] ? this._customSize[item._listId] : this._itemSize.width;
        }
        if (
            (!this._sizeType && item.width != defaultWidth)
            || (this._sizeType && item.height != defaultHeight)
        ) {
            if (!this._customSize)
                this._customSize = {};
            let val = this._sizeType ? item.height : item.width;
            if (this._customSize[item._listId] != val) {
                const beforeSize = this._customSize[item._listId] ? this._customSize[item._listId] : (this._sizeType ? this._itemSize.height : this._itemSize.width);
                this._customSize[item._listId] = val;
                if (this.adaptiveSize) {
                    const sizeChange = val - beforeSize;
                    this.resetItemPos(item, sizeChange);
                }
                this._resizeContent();
                // this.content.children.forEach((child: cc.Node) => {
                //     this._updateItemPos(child);
                // });
                this.updateAll();
                // 如果当前正在运行 scrollTo，肯定会不准确，在这里做修正
                if (this._scrollToListId != null) {
                    this._scrollPos = null;
                    this.unschedule(this._scrollToSo);
                    this.scrollTo(this._scrollToListId, Math.max(0, this._scrollToEndTime - ((new Date()).getTime() / 1000)));
                }
            }
        }
        // }
    }

    private resetItemPos(item: cc.Node, sizeChange: number): void {
        const id = item._listId + 1;
        this._fixDelta += sizeChange;
        for (let index = id; index < this._customPos.length; ++index) {
            this._customPos[index] -= sizeChange;
        }
    }


    //PAGE粘附
    _pageAdhere() {
        if (!this.cyclic && (this.elasticTop > 0 || this.elasticRight > 0 || this.elasticBottom > 0 || this.elasticLeft > 0))
            return;
        let curPos = this._sizeType ? this.viewTop : this.viewLeft;
        let dis = (this._sizeType ? this.node.height : this.node.width) * this.pageDistance;
        let canSkip = Math.abs(this._beganPos - curPos) > dis;
        if (canSkip) {
            let timeInSecond = .5;
            switch (this._alignCalcType) {
                case 1://单行HORIZONTAL（LEFT_TO_RIGHT）、网格VERTICAL（LEFT_TO_RIGHT）
                case 4://单列VERTICAL（BOTTOM_TO_TOP）、网格HORIZONTAL（BOTTOM_TO_TOP）
                    if (this._beganPos > curPos) {
                        this.prePage(timeInSecond);
                        // cc.log('_pageAdhere   PPPPPPPPPPPPPPP');
                    } else {
                        this.nextPage(timeInSecond);
                        // cc.log('_pageAdhere   NNNNNNNNNNNNNNN');
                    }
                    break;
                case 2://单行HORIZONTAL（RIGHT_TO_LEFT）、网格VERTICAL（RIGHT_TO_LEFT）
                case 3://单列VERTICAL（TOP_TO_BOTTOM）、网格HORIZONTAL（TOP_TO_BOTTOM）
                    if (this._beganPos < curPos) {
                        this.prePage(timeInSecond);
                    } else {
                        this.nextPage(timeInSecond);
                    }
                    break;
            }
        } else if (this.elasticTop <= 0 && this.elasticRight <= 0 && this.elasticBottom <= 0 && this.elasticLeft <= 0) {
            this.adhere();
        }
        this._beganPos = null;
    }
    //粘附
    adhere() {
        if (!this.checkInited())
            return;
        if (this.elasticTop > 0 || this.elasticRight > 0 || this.elasticBottom > 0 || this.elasticLeft > 0)
            return;
        this.adhering = true;
        this._calcNearestItem();
        let offset: number = (this._sizeType ? this._topGap : this._leftGap) / (this._sizeType ? this.node.height : this.node.width);
        let timeInSecond: number = .7;
        this.scrollTo(this.nearestListId, timeInSecond, offset);
    }
    //Update..
    update() {
        if (this.frameByFrameRenderNum <= 0 || this._updateDone)
            return;
        // cc.log(this.displayData.length, this._updateCounter, this.displayData[this._updateCounter]);
        if (this._virtual) {
            let len: number = (this._updateCounter + this.frameByFrameRenderNum) > this.displayItemNum ? this.displayItemNum : (this._updateCounter + this.frameByFrameRenderNum);
            for (let n: number = this._updateCounter; n < len; n++) {
                let data: any = this.displayData[n];
                if (data) {
                    this._createOrUpdateItem(data);
                }
            }

            if (this._updateCounter >= this.displayItemNum - 1) { //最后一个
                if (this._doneAfterUpdate) {
                    this._updateCounter = 0;
                    this._updateDone = false;
                    // if (!this._scrollView.isScrolling())
                    this._doneAfterUpdate = false;
                } else {
                    this._updateDone = true;
                    this._delRedundantItem();
                    this._forceUpdate = false;
                    this._calcNearestItem();
                    if (this.slideMode == SlideType.PAGE)
                        this.curPageNum = this.nearestListId;
                }
            } else {
                this._updateCounter += this.frameByFrameRenderNum;
            }
        } else {
            if (this._updateCounter < this._numItems) {
                let len: number = (this._updateCounter + this.frameByFrameRenderNum) > this._numItems ? this._numItems : (this._updateCounter + this.frameByFrameRenderNum);
                for (let n: number = this._updateCounter; n < len; n++) {
                    this._createOrUpdateItem2(n);
                }
                this._updateCounter += this.frameByFrameRenderNum;
            } else {
                this._updateDone = true;
                this._calcNearestItem();
                if (this.slideMode == SlideType.PAGE)
                    this.curPageNum = this.nearestListId;
            }
        }
    }
    /**
     * 创建或更新Item（虚拟列表用）
     * @param {Object} data 数据
     */
    _createOrUpdateItem(data: any) {
        let item: any = this.getItemByListId(data.id);
        if (!item) { //如果不存在
            let canGet: boolean = this._pool.size() > 0;
            if (canGet) {
                item = this._pool.get();
                // cc.log('从池中取出::   旧id =', item['_listId'], '，新id =', data.id, item);
            } else {
                item = cc.instantiate(this._itemTmp);
                // cc.log('新建::', data.id, item);
            }
            if (!canGet || !cc.isValid(item)) {
                item = cc.instantiate(this._itemTmp);
                canGet = false;
            }
            if (item._listId != data.id) {
                item._listId = data.id;
                // item.setContentSize(this._itemSize);
            }
            item.setPosition(cc.v2(data.x, data.y));
            this._resetItemSize(item);
            this.content.addChild(item);

            if (canGet && this._needUpdateWidget) {
                let widget: cc.Widget = item.getComponent(cc.Widget);
                if (widget)
                    widget.updateAlignment();
            }
            item.setSiblingIndex(this.content.childrenCount - 1);

            let listItem: ListItem = item.getComponent(this.itemComponentName);
            item['listItem'] = listItem;
            if (listItem) {
                listItem.listId = data.id;
                listItem.list = this;
                listItem._registerEvent();
            }

            if (this.renderEvent) {
                cc.Component.EventHandler.emitEvents([this.renderEvent], item, data.id % this._actualNumItems);
            }

        } else if (this._forceUpdate && this.renderEvent) { //强制更新
            item.setPosition(cc.v2(data.x, data.y));
            this._resetItemSize(item);
            // cc.log('ADD::', data.id, item);
            if (this.renderEvent) {
                cc.Component.EventHandler.emitEvents([this.renderEvent], item, data.id % this._actualNumItems);
            }
        }
        // this._resetItemSize(item);

        this._updateListItem(item['listItem']);
        if (this._lastDisplayData.indexOf(data.id) < 0) {
            this._lastDisplayData.push(data.id);
        }
    }
    //创建或更新Item（非虚拟列表用）
    _createOrUpdateItem2(listId: number) {
        let item: any = this.content.children[listId];
        let listItem: ListItem;
        if (!item) { //如果不存在
            item = cc.instantiate(this._itemTmp);
            item._listId = listId;
            this.content.addChild(item);

            listItem = item.getComponent(this.itemComponentName);
            item['listItem'] = listItem;
            if (listItem) {
                listItem.listId = listId;
                listItem.list = this;
                listItem._registerEvent();
            }
            if (this.renderEvent) {
                cc.Component.EventHandler.emitEvents([this.renderEvent], item, listId % this._actualNumItems);
            }
        } else if (this._forceUpdate && this.renderEvent) { //强制更新
            item._listId = listId;
            if (listItem)
                listItem.listId = listId;
            if (this.renderEvent) {
                cc.Component.EventHandler.emitEvents([this.renderEvent], item, listId % this._actualNumItems);
            }
        }
        this._updateListItem(listItem);
        if (this._lastDisplayData.indexOf(listId) < 0) {
            this._lastDisplayData.push(listId);
        }
    }

    _updateListItem(listItem: ListItem) {
        if (!listItem)
            return;
        if (this.selectedMode > SelectedType.NONE) {
            let item: any = listItem.node;
            switch (this.selectedMode) {
                case SelectedType.SINGLE:
                    listItem.selected = this.selectedId == item._listId;
                    break;
                case SelectedType.MULT:
                    listItem.selected = this.multSelected.indexOf(item._listId) >= 0;
                    break;
            }
        }
    }
    //仅虚拟列表用
    _resetItemSize(item: any) {
        // return;
        let size: number;
        if (this._customSize && this._customSize[item._listId]) {
            size = this._customSize[item._listId];
        } else {
            if (this._colLineNum > 1)
                item.setContentSize(this._itemSize);
            else
                size = this._sizeType ? this._itemSize.height : this._itemSize.width;
        }
        if (size) {
            if (this._sizeType)
                item.height = size;
            else
                item.width = size;
        }
    }
    /**
     * 更新Item位置
     * @param {Number||Node} listIdOrItem
     */
    _updateItemPos(listIdOrItem: any) {
        let item: any = isNaN(listIdOrItem) ? listIdOrItem : this.getItemByListId(listIdOrItem);
        let pos: any = this.getItemPos(item._listId);
        item.setPosition(pos.x, pos.y);
    }
    /**
     * 设置多选
     * @param {Array} args 可以是单个listId，也可是个listId数组
     * @param {Boolean} bool 值，如果为null的话，则直接用args覆盖
     */
    setMultSelected(args: any, bool: boolean) {
        if (!this.checkInited())
            return;
        if (!Array.isArray(args)) {
            args = [args];
        }
        if (bool == null) {
            this.multSelected = args;
        } else {
            let listId: number, sub: number;
            if (bool) {
                for (let n: number = args.length - 1; n >= 0; n--) {
                    listId = args[n];
                    sub = this.multSelected.indexOf(listId);
                    if (sub < 0) {
                        this.multSelected.push(listId);
                    }
                }
            } else {
                for (let n: number = args.length - 1; n >= 0; n--) {
                    listId = args[n];
                    sub = this.multSelected.indexOf(listId);
                    if (sub >= 0) {
                        this.multSelected.splice(sub, 1);
                    }
                }
            }
        }
        this._forceUpdate = true;
        this._onScrolling();
    }
    /**
     * 获取多选数据
     * @returns
     */
    getMultSelected() {
        return this.multSelected;
    }
    /**
     * 多选是否有选择
     * @param {number} listId 索引
     * @returns
     */
    hasMultSelected(listId: number) {
        return this.multSelected && this.multSelected.indexOf(listId) >= 0;
    }
    /**
     * 更新指定的Item
     * @param {Array} args 单个listId，或者数组
     * @returns
     */
    updateItem(args: any) {
        if (!this.checkInited())
            return;
        if (!Array.isArray(args)) {
            args = [args];
        }
        for (let n: number = 0, len: number = args.length; n < len; n++) {
            let listId: number = args[n];
            let item: any = this.getItemByListId(listId);
            if (item)
                cc.Component.EventHandler.emitEvents([this.renderEvent], item, listId % this._actualNumItems);
        }
    }
    /**
     * 更新全部
     */
    updateAll() {
        if (!this.checkInited())
            return;
        this.numItems = this.numItems;
    }
    /**
     * 根据ListID获取Item
     * @param {Number} listId
     * @returns
     */
    getItemByListId(listId: number) {
        if (this.content) {
            for (let n: number = this.content.childrenCount - 1; n >= 0; n--) {
                let item: any = this.content.children[n];
                if (item._listId == listId)
                    return item;
            }
        }
    }
    /**
     * 获取在显示区域外的Item
     * @returns
     */
    _getOutsideItem() {
        let item: any;
        let result: any[] = [];
        for (let n: number = this.content.childrenCount - 1; n >= 0; n--) {
            item = this.content.children[n];
            if (!this.displayData.find(d => d.id == item._listId)) {
                result.push(item);
            }
        }
        return result;
    }
    //删除显示区域以外的Item
    _delRedundantItem() {
        if (this._virtual) {
            let arr: any[] = this._getOutsideItem();
            for (let n: number = arr.length - 1; n >= 0; n--) {
                let item: any = arr[n];
                if (this._scrollItem && item._listId == this._scrollItem._listId)
                    continue;
                // 2023/8/18 cft listItem复用的时候文字会被清空
                // let comLabels = item.getComponentsInChildren(cc.Label);
                // let comRichTexts = item.getComponentsInChildren(cc.RichText);
                // if (comLabels) {
                //     for (let index = 0; index < comLabels.length; index++) {
                //         let element = comLabels[index];
                //         element.string = ""
                //         element.oldstring = ""
                //     }
                // }
                // if (comRichTexts) {
                //     for (let index = 0; index < comRichTexts.length; index++) {
                //         let element = comRichTexts[index];
                //         element.string = ""
                //         element.oldstring = ""
                //     }
                // }
                item.isCached = true;
                this._pool.put(item);
                for (let m: number = this._lastDisplayData.length - 1; m >= 0; m--) {
                    if (this._lastDisplayData[m] == item._listId) {
                        this._lastDisplayData.splice(m, 1);
                        break;
                    }
                }
            }
            // cc.log('存入::', str, '    pool.length =', this._pool.length);
        } else {
            while (this.content.childrenCount > this._numItems) {
                this._delSingleItem(this.content.children[this.content.childrenCount - 1]);
            }
        }
    }
    //删除单个Item
    _delSingleItem(item: any) {
        // cc.log('DEL::', item['_listId'], item);
        item.removeFromParent();
        if (item.destroy)
            item.destroy();
        item = null;
    }
    /** 
     * 动效删除Item（此方法只适用于虚拟列表，即_virtual=true）
     * 一定要在回调函数里重新设置新的numItems进行刷新，毕竟本List是靠数据驱动的。
     */
    aniDelItem(listId: number, callFunc: Function, aniType: number) {
        if (!this.checkInited() || this.cyclic || !this._virtual)
            return cc.error('This function is not allowed to be called!');

        if (!callFunc)
            return cc.error('CallFunc are not allowed to be NULL, You need to delete the corresponding index in the data array in the CallFunc!');

        if (this._aniDelRuning)
            return cc.warn('Please wait for the current deletion to finish!');

        let item: any = this.getItemByListId(listId);
        let listItem: ListItem;
        if (!item) {
            callFunc(listId);
            return;
        } else {
            listItem = item.getComponent(this.itemComponentName);
        }
        this._aniDelRuning = true;
        this._aniDelCB = callFunc;
        this._aniDelItem = item;
        this._aniDelBeforePos = item.position;
        this._aniDelBeforeScale = item.scale;
        let curLastId: number = this.displayData[this.displayData.length - 1].id;
        let resetSelectedId: boolean = listItem.selected;
        listItem.showAni(aniType, () => {
            //判断有没有下一个，如果有的话，创建粗来
            let newId: number;
            if (curLastId < this._numItems - 2) {
                newId = curLastId + 1;
            }
            if (newId != null) {
                let newData: any = this._calcItemPos(newId);
                this.displayData.push(newData);
                if (this._virtual)
                    this._createOrUpdateItem(newData);
                else
                    this._createOrUpdateItem2(newId);
            } else
                this._numItems--;
            if (this.selectedMode == SelectedType.SINGLE) {
                if (resetSelectedId) {
                    this._selectedId = -1;
                } else if (this._selectedId - 1 >= 0) {
                    this._selectedId--;
                }
            } else if (this.selectedMode == SelectedType.MULT && this.multSelected.length) {
                let sub: number = this.multSelected.indexOf(listId);
                if (sub >= 0) {
                    this.multSelected.splice(sub, 1);
                }
                //多选的数据，在其后的全部减一
                for (let n: number = this.multSelected.length - 1; n >= 0; n--) {
                    let id: number = this.multSelected[n];
                    if (id >= listId)
                        this.multSelected[n]--;
                }
            }
            if (this._customSize) {
                if (this._customSize[listId])
                    delete this._customSize[listId];
                let newCustomSize: any = {};
                let size: number;
                for (let id in this._customSize) {
                    size = this._customSize[id];
                    let idNumber: number = parseInt(id);
                    newCustomSize[idNumber - (idNumber >= listId ? 1 : 0)] = size;
                }
                this._customSize = newCustomSize;
            }
            //后面的Item向前怼的动效
            let sec: number = .2333;
            let tween: cc.Tween, haveCB: boolean;
            for (let n: number = newId != null ? newId : curLastId; n >= listId + 1; n--) {
                item = this.getItemByListId(n);
                if (item) {
                    let posData: any = this._calcItemPos(n - 1);
                    tween = cc.tween(item)
                        .to(sec, { position: cc.v2(posData.x, posData.y) });
                    if (n <= listId + 1) {
                        haveCB = true;
                        tween.call(() => {
                            this._aniDelRuning = false;
                            callFunc(listId);
                            delete this._aniDelCB;
                        });
                    }
                    tween.start();
                }
            }
            if (!haveCB) {
                this._aniDelRuning = false;
                callFunc(listId);
                this._aniDelCB = null;
            }
        }, true);
    }
    /**
     * 滚动到..
     * @param {Number} listId 索引（如果<0，则滚到首个Item位置，如果>=_numItems，则滚到最末Item位置）
     * @param {Number} timeInSecond 时间
     * @param {Number} offset 索引目标位置偏移，0-1
     * @param {Boolean} overStress 滚动后是否强调该Item（这只是个实验功能）
     */
    scrollTo(listId: number, timeInSecond: number = .5, offset: number = null, overStress: boolean = false) {
        if (!this.checkInited(false))
            return;
        // t._scrollView.stopAutoScroll();
        if (timeInSecond == null)   //默认0.5
            timeInSecond = .5;
        else if (timeInSecond < 0)
            timeInSecond = 0;
        if (listId < 0)
            listId = 0;
        else if (listId >= this._numItems)
            listId = this._numItems - 1;
        // 以防设置了numItems之后layout的尺寸还未更新
        if (!this._virtual && this._layout && this._layout.enabled)
            this._layout.updateLayout();

        let pos = this.getItemPos(listId);
        if (!pos) {
            return CC_DEV && cc.error('pos is null', listId);
        }
        let targetX: number, targetY: number;

        switch (this._alignCalcType) {
            case 1://单行HORIZONTAL（LEFT_TO_RIGHT）、网格VERTICAL（LEFT_TO_RIGHT）
                targetX = pos.left;
                if (offset != null)
                    targetX -= this.node.width * offset;
                else
                    targetX -= this._leftGap;
                pos = cc.v2(targetX, 0);
                break;
            case 2://单行HORIZONTAL（RIGHT_TO_LEFT）、网格VERTICAL（RIGHT_TO_LEFT）
                targetX = pos.right - this.node.width;
                if (offset != null)
                    targetX += this.node.width * offset;
                else
                    targetX += this._rightGap;
                pos = cc.v2(targetX + this.content.width, 0);
                break;
            case 3://单列VERTICAL（TOP_TO_BOTTOM）、网格HORIZONTAL（TOP_TO_BOTTOM）
                targetY = pos.top;
                if (offset != null)
                    targetY += this.node.height * offset;
                else
                    targetY += this._topGap;
                pos = cc.v2(0, -targetY);
                break;
            case 4://单列VERTICAL（BOTTOM_TO_TOP）、网格HORIZONTAL（BOTTOM_TO_TOP）
                targetY = pos.bottom + this.node.height;
                if (offset != null)
                    targetY -= this.node.height * offset;
                else
                    targetY -= this._bottomGap;
                pos = cc.v2(0, -targetY + this.content.height);
                break;
        }
        let viewPos: any = this.content.getPosition();
        viewPos = Math.abs(this._sizeType ? viewPos.y : viewPos.x);

        let comparePos = this._sizeType ? pos.y : pos.x;
        let runScroll = Math.abs((this._scrollPos != null ? this._scrollPos : viewPos) - comparePos) > .5;
        // cc.log(runScroll, t._scrollPos, viewPos, comparePos)

        // t._scrollView.stopAutoScroll();
        if (runScroll) {
            this._scrollView.scrollToOffset(pos, timeInSecond);
            this._scrollToListId = listId;
            this._scrollToEndTime = ((new Date()).getTime() / 1000) + timeInSecond;
            // cc.log(listId, t.content.width, t.content.getPosition(), pos);
            this._scrollToSo = this.scheduleOnce(() => {
                if (!this._adheringBarrier) {
                    this.adhering = this._adheringBarrier = false;
                }
                this._scrollPos =
                    this._scrollToListId =
                    this._scrollToEndTime =
                    this._scrollToSo =
                    null;
                //cc.log('2222222222', t._adheringBarrier)
                if (overStress) {
                    // t.scrollToListId = listId;
                    let item = this.getItemByListId(listId);
                    if (item) {
                        cc.tween(item)
                            .to(.1, { scale: 1.05 })
                            .to(.1, { scale: 1 })
                            .start();
                    }
                }
            }, timeInSecond + .1);

            if (timeInSecond <= 0) {
                this._onScrolling();
            }
        }
    }
    /**
     * 计算当前滚动窗最近的Item
     */
    _calcNearestItem() {
        this.nearestListId = null;
        let data: any, center: number;

        if (this._virtual)
            this._calcViewPos();

        let vTop: number, vRight: number, vBottom: number, vLeft: number;
        vTop = this.viewTop;
        vRight = this.viewRight;
        vBottom = this.viewBottom;
        vLeft = this.viewLeft;

        let breakFor: boolean = false;
        for (let n = 0; n < this.content.childrenCount && !breakFor; n += this._colLineNum) {
            data = this._virtual ? this.displayData[n] : this._calcExistItemPos(n);
            if (data) {
                center = this._sizeType ? ((data.top + data.bottom) / 2) : (center = (data.left + data.right) / 2);
                switch (this._alignCalcType) {
                    case 1://单行HORIZONTAL（LEFT_TO_RIGHT）、网格VERTICAL（LEFT_TO_RIGHT）
                        if (data.right >= vLeft) {
                            this.nearestListId = data.id;
                            if (vLeft > center)
                                this.nearestListId += this._colLineNum;
                            breakFor = true;
                        }
                        break;
                    case 2://单行HORIZONTAL（RIGHT_TO_LEFT）、网格VERTICAL（RIGHT_TO_LEFT）
                        if (data.left <= vRight) {
                            this.nearestListId = data.id;
                            if (vRight < center)
                                this.nearestListId += this._colLineNum;
                            breakFor = true;
                        }
                        break;
                    case 3://单列VERTICAL（TOP_TO_BOTTOM）、网格HORIZONTAL（TOP_TO_BOTTOM）
                        if (data.bottom <= vTop) {
                            this.nearestListId = data.id;
                            if (vTop < center)
                                this.nearestListId += this._colLineNum;
                            breakFor = true;
                        }
                        break;
                    case 4://单列VERTICAL（BOTTOM_TO_TOP）、网格HORIZONTAL（BOTTOM_TO_TOP）
                        if (data.top >= vBottom) {
                            this.nearestListId = data.id;
                            if (vBottom > center)
                                this.nearestListId += this._colLineNum;
                            breakFor = true;
                        }
                        break;
                }
            }
        }
        //判断最后一个Item。。。（哎，这些判断真心恶心，判断了前面的还要判断最后一个。。。一开始呢，就只有一个布局（单列布局），那时候代码才三百行，后来就想着完善啊，艹..这坑真深，现在这行数都一千五了= =||）
        data = this._virtual ? this.displayData[this.displayItemNum - 1] : this._calcExistItemPos(this._numItems - 1);
        if (data && data.id == this._numItems - 1) {
            center = this._sizeType ? ((data.top + data.bottom) / 2) : (center = (data.left + data.right) / 2);
            switch (this._alignCalcType) {
                case 1://单行HORIZONTAL（LEFT_TO_RIGHT）、网格VERTICAL（LEFT_TO_RIGHT）
                    if (vRight > center)
                        this.nearestListId = data.id;
                    break;
                case 2://单行HORIZONTAL（RIGHT_TO_LEFT）、网格VERTICAL（RIGHT_TO_LEFT）
                    if (vLeft < center)
                        this.nearestListId = data.id;
                    break;
                case 3://单列VERTICAL（TOP_TO_BOTTOM）、网格HORIZONTAL（TOP_TO_BOTTOM）
                    if (vBottom < center)
                        this.nearestListId = data.id;
                    break;
                case 4://单列VERTICAL（BOTTOM_TO_TOP）、网格HORIZONTAL（BOTTOM_TO_TOP）
                    if (vTop > center)
                        this.nearestListId = data.id;
                    break;
            }
        }
        // cc.log('t.nearestListId =', t.nearestListId);
    }
    //上一页
    prePage(timeInSecond: number = .5) {
        // cc.log('👈');
        if (!this.checkInited())
            return;
        this.skipPage(this.curPageNum - 1, timeInSecond);
    }
    //下一页
    nextPage(timeInSecond: number = .5) {
        // cc.log('👉');
        if (!this.checkInited())
            return;
        this.skipPage(this.curPageNum + 1, timeInSecond);
    }
    //跳转到第几页
    skipPage(pageNum: number, timeInSecond: number) {
        if (!this.checkInited())
            return;
        if (this._slideMode != SlideType.PAGE)
            return cc.error('This function is not allowed to be called, Must SlideMode = PAGE!');
        if (pageNum < 0 || pageNum >= this._numItems)
            return;
        if (this.curPageNum == pageNum)
            return;
        // cc.log(pageNum);
        this.curPageNum = pageNum;
        if (this.pageChangeEvent) {
            cc.Component.EventHandler.emitEvents([this.pageChangeEvent], pageNum);
        }
        this.scrollTo(pageNum, timeInSecond);
    }
    //计算 CustomSize（这个函数还是保留吧，某些罕见的情况的确还是需要手动计算customSize的）
    calcCustomSize(numItems: number) {
        if (!this.checkInited())
            return;
        if (!this._itemTmp)
            return cc.error('Unset template item!');
        if (!this.renderEvent)
            return cc.error('Unset Render-Event!');
        this._customSize = {};
        let temp: any = cc.instantiate(this._itemTmp);
        this.content.addChild(temp);
        for (let n: number = 0; n < numItems; n++) {
            cc.Component.EventHandler.emitEvents([this.renderEvent], temp, n);
            if (temp.height != this._itemSize.height || temp.width != this._itemSize.width) {
                this._customSize[n] = this._sizeType ? temp.height : temp.width;
            }
        }
        if (!Object.keys(this._customSize).length)
            this._customSize = null;
        temp.removeFromParent();
        if (temp.destroy)
            temp.destroy();
        return this._customSize;
    }
}